 /*----------------------------------------------------------------------*
    smpl11.c
    Tv11 - Version 1.0
    [DynamixelvgR]
    Copyright (C) 2009 BestTechnology CO.,LTD.
 *----------------------------------------------------------------------*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <util/delay.h>
#include <avr/wdt.h>
#include <stdbool.h>
#include <string.h>
#include <rs_hp.h>             // USARTʐMCu

#define rbi(sfr, bit) (_SFR_BYTE(sfr) ^= _BV(bit))

#define MAX_DISTANCE (999)

#define BROADCASTING_ID (0xfeU)

// CXgNVpPbgXe[g
typedef enum {
  stIdle, stHeader, stLoad, stSum
} TPacketStatus;

// G[rbg
typedef enum {
  erInst=_BV(6), erOverload=_BV(5), erCheckSum=_BV(4), erRange=_BV(3), erOverheat=_BV(2), erAngelLimit=_BV(1), erInputVoltage=_BV(0), erNon=0
} TDxError;

// CXgNV
typedef enum {
  INST_PING       = 1,
  INST_READ       = 2,
  INST_WRITE      = 3,
  INST_REG_WRITE  = 4,
  INST_ACTION     = 5,
  INST_RESET      = 6,
  INST_SYS_WRITE  = 0xD,
  INST_SYNC_WRITE = 0x83,
  INST_SYNG_REG_WRITE = 0x85
} TInstruction;

// memory map (fixed data)
typedef struct {
  // 0(0x00)
  uint8_t ModelNumberL;
  // 1(0x01)
  uint8_t ModelNumberH;
  // 2(0x02)
  uint8_t FirmwareVersion;
} TFixROM;

typedef struct {  // 11byte
  // 0
  uint8_t   PulseNum;
  // 1
  uint16_t  CYCLE;
  // 3
  uint16_t  Deadband;
  // 5
  uint16_t  Threshold;
  // 7
  uint16_t  DAC;
  // 9
  uint16_t  DetectLevel;
  // 11
  uint16_t  SpeedOfSound;
} TVariableParam;

typedef struct {
  // 3(0x03)
  uint8_t   ID;
  // 4(0x04)
  uint8_t   BaudRate;
  // 5(0x05)
  uint8_t   ReturnDelayTime;
  // 6(0x06)
  uint8_t   MODE;
  // 7(0x07)
  TVariableParam  VariableParam;
  // 20(0x14)
  uint8_t   StatusReturnLevel;
  // 19(0x15)
  uint8_t   reserve[4];
  // 24~(0x18)
} TVariableROM;

typedef struct {
  // 25(0x18)
  uint8_t   Led;
  // 26(0x1A)
  uint8_t   reserve2;
  // 27(0x1B)
  TVariableParam  VariableParam;
  // 40(0x28)
  uint8_t   reserve3;
  // 41(0x29)
  uint8_t   SCANSW;
  // 42(0x2A)
  uint16_t  Distance;   // (W) v
  // 44(0x2C)..52(0x34)
  uint16_t  Detect[5];  // (W) o
} TVariableRAM;

typedef struct {
  TFixROM       FixROM;
  TVariableROM  ROM ;
} TMemoryMap_EEPROM;

typedef union {
  struct {
    TFixROM       FixROM;
    TVariableROM  VariableROM ;
    TVariableRAM  VariableRAM ;
  };
  uint8_t BYTE[sizeof(TFixROM) + sizeof(TVariableROM) + sizeof(TVariableRAM)];
} TMemoryMap;

TMemoryMap MemoryMap, NewMemoryMap;

// factory default
const TFixROM DefaultFixROM = {
  0x4, 0x40, 'K'
};

const TVariableROM DefaultVariableROM = {
  0, 34, 80,
  1,
  VariableParam:{
    5,    // PulseNum
    4000, // CYCLE
    100,  // Deadband
    150,  // Threshold
    300,  // DAC
    768,  // DetectLevel
    343   // SpeedOfSound
  },
  2
};

const TMemoryMap_EEPROM DEFAULT_DATA /*__attribute__((section(".eeprom")))*/ = {
  FixROM:{
    0x4, 0x40, 'K'
  },
  ROM:{
    0, 1/*34*/, 80,
    1,
    VariableParam:{
      5,    // PulseNum
      4000, // CYCLE
      100,  // Deadband
      150,  // Threshold
      300,  // DAC
      768,  // DetectLevel
      343   // SpeedOfSound
    },
    2
  }
};

volatile uint8_t  TimeOut;                    // ʐM^CAEg
uint16_t          MesTime_local = 0;
uint16_t          MesTime;
bool              MesTimeFlag = true;
uint16_t          oldMesTime = 0;
uint8_t           DetectCount = 0;            // o/Cycle
uint8_t           PulseCount = 0;             // go/Cycle
volatile uint32_t ElapsedTime;                // NԌvpJE^
volatile bool     ElapsedTimeFlag = true;
volatile bool     ExclusionMesWrite = false;  // r
volatile bool     ExclusionDAWrite = false;
uint8_t           packet[128];                // MpPbgf[^

/*-----------------------------------------------*/
/* EFCg[` (^C}/JE^2gp)     */
/*-----------------------------------------------*/
void wait (unsigned long pause_ms)
{
  unsigned long i;

  TCCR2A = 0x00;
  TCCR2B = 0x00;   // stop
  TCNT2  = 0;      // set count
  OCR2A  = 0xfa;   // set compare
  TCCR2A = 0x02;   // ctc
  TCCR2B = 0x04;   // start timer clk/64  (OCF=976.5625Hz1ms)

  for (i = 0; i < pause_ms; i++) {
    loop_until_bit_is_set(TIFR2, OCF2A);
    TIFR2 |= _BV(OCF2A);
  }
  TCCR2B = 0x00;   // stop
}

void InitIO (void) {
  // [hؑփXCb`(PD2)\
  DDRD &= ~_BV(DDD2); // PD2͂
  PORTD |= _BV(PD2);  // PD2̓vAbvRON

  // j^LED1(PB0)\
  DDRB |= _BV(DDB0);  // PBo͂
  PORTB |= _BV(PB0);  // LEDOFF PB0=0

  // j^LED2(PD3)\
  DDRD |= _BV(DDD3);  // PDo͂
  PORTD |= _BV(PD3);  // LEDOFF PD3=0

  // j^LED3(PD7)\
  DDRD |= _BV(DDD7);  // PDo͂
  PORTD |= _BV(PD7);  // LEDOFF PD7=0

  // ADC
  ADCSRA = _BV(ADEN) | _BV(ADIE) | _BV(ADPS2);  // A/Dϊ, ϊNbNCK/16)

  // AC
  DIDR1 = 0x03;           // Rp[^[q̃fW^͋֎~
  ACSR = 0x13;            // Up EdgeA݋֎~

  eeprom_busy_wait ();
//  eeprom_read_block (&MemoryMap, (void *)0, sizeof(TFixROM) + sizeof(TVariableROM));
  memcpy (&MemoryMap, &DEFAULT_DATA, sizeof(TFixROM) + sizeof(TVariableROM));
  memcpy (&MemoryMap.VariableRAM.VariableParam, &MemoryMap.VariableROM.VariableParam, sizeof(TVariableParam));

  eeprom_busy_wait ();
  if (eeprom_read_word ((void *)510) == 0xDEAD) eeprom_write_word ((void *)510, 0xffff);
}

void Led1 (bool b) {
  if (b) {
    PORTB &= ~_BV(PB0);       // LED ON
  } else {
    PORTB |= _BV(PB0);        // LED OFF
  }
}

void Led2 (bool b) {
  if (b) {
    PORTD &= ~_BV(PD3);       // LED ON
  } else {
    PORTD |= _BV(PD3);        // LED OFF
  }
}

void Led3 (bool b) {
  if (b) {
    PORTD &= ~_BV(PD7);       // LED ON
  } else {
    PORTD |= _BV(PD7);        // LED OFF
  }
}

/*======================================================================*/
/* pPbg                                                         */
/*======================================================================*/
// wb_M
unsigned char PutHeader (unsigned char id, unsigned char len, TDxError err) {
  _delay_loop_2 ((((unsigned short)MemoryMap.VariableROM.ReturnDelayTime) << 3) + 1);
  rs_hp_changedir (true);
  _delay_loop_1 (10);
  rs_hp_putc (0xff); rs_hp_putc (0xff);
  rs_hp_putc (id); rs_hp_putc (len); rs_hp_putc (err);
  return id + len + err;
}

// TM
void PutSum (uint8_t sum) {
  rs_hp_putc (sum);
  rs_hp_changedir (false);
}

/*-----------------------------------------------*/
/* C                                        */
/*-----------------------------------------------*/
void main(void)
{
  #define limit(d,max,min)  (((d)<=(max))&&((d)>=(min)))

  void memwrite(uint8_t oldid, bool bcast) {
    uint8_t csum;
    bool changebaud, changeda, changerom;

    // range check
    if (
      (NewMemoryMap.VariableROM.ID <= (uint8_t)253)
      && (NewMemoryMap.VariableROM.BaudRate <= (uint8_t)254)
      && (NewMemoryMap.VariableROM.ReturnDelayTime <= (uint8_t)254)
      && (NewMemoryMap.VariableROM.StatusReturnLevel <= (uint8_t)2)

      && limit(NewMemoryMap.VariableROM.VariableParam.PulseNum, (uint8_t)250, (uint8_t)2)
      && limit(NewMemoryMap.VariableROM.VariableParam.CYCLE, 8000, 200)
      && limit(NewMemoryMap.VariableROM.VariableParam.Deadband, 8000, 0)
      && limit(NewMemoryMap.VariableROM.VariableParam.Threshold, 1000, 0)
      && limit(NewMemoryMap.VariableROM.VariableParam.DAC, 4095, 1)
      && limit(NewMemoryMap.VariableROM.VariableParam.DetectLevel, 4095, 0)

      && limit(NewMemoryMap.VariableRAM.VariableParam.PulseNum, (uint8_t)250, (uint8_t)2)
      && limit(NewMemoryMap.VariableRAM.VariableParam.CYCLE, 8000, 200)
      && limit(NewMemoryMap.VariableRAM.VariableParam.Deadband, 8000, 0)
      && limit(NewMemoryMap.VariableRAM.VariableParam.Threshold, 1000, 0)
      && limit(NewMemoryMap.VariableRAM.VariableParam.DAC, 4095, 0)
      && limit(NewMemoryMap.VariableRAM.VariableParam.DetectLevel, 4095, 0)

      && limit(NewMemoryMap.VariableRAM.VariableParam.SpeedOfSound, 999, 0)
    ) {
      changebaud =
        (MemoryMap.VariableROM.BaudRate != NewMemoryMap.VariableROM.BaudRate);
      changeda =
        (MemoryMap.VariableRAM.VariableParam.DetectLevel != NewMemoryMap.VariableRAM.VariableParam.DetectLevel);
      changerom = (
        (MemoryMap.VariableROM.ID != NewMemoryMap.VariableROM.ID)
        || changebaud
        || (MemoryMap.VariableROM.MODE != NewMemoryMap.VariableROM.MODE)
        || (MemoryMap.VariableROM.StatusReturnLevel != NewMemoryMap.VariableROM.StatusReturnLevel)
        || (MemoryMap.VariableROM.ReturnDelayTime != NewMemoryMap.VariableROM.ReturnDelayTime)
        || (MemoryMap.VariableROM.VariableParam.PulseNum != NewMemoryMap.VariableROM.VariableParam.PulseNum)
        || (MemoryMap.VariableROM.VariableParam.CYCLE != NewMemoryMap.VariableROM.VariableParam.CYCLE)
        || (MemoryMap.VariableROM.VariableParam.Deadband != NewMemoryMap.VariableROM.VariableParam.Deadband)
        || (MemoryMap.VariableROM.VariableParam.Threshold != NewMemoryMap.VariableROM.VariableParam.Threshold)
        || (MemoryMap.VariableROM.VariableParam.DAC != NewMemoryMap.VariableROM.VariableParam.DAC)
        || (MemoryMap.VariableROM.VariableParam.DetectLevel != NewMemoryMap.VariableROM.VariableParam.DetectLevel)
        || (MemoryMap.VariableROM.VariableParam.SpeedOfSound != NewMemoryMap.VariableROM.VariableParam.SpeedOfSound)
      );

      memcpy (&MemoryMap, &NewMemoryMap, sizeof(TMemoryMap));
      if (changerom) {
        eeprom_busy_wait ();
        eeprom_write_block (&MemoryMap.VariableROM, (uint8_t *)sizeof(TFixROM), sizeof(TVariableROM));
      }

      // success
      if ((MemoryMap.VariableROM.StatusReturnLevel == 2) && !bcast) {
        csum = PutHeader (oldid, 2, erNon);
        PutSum (~csum);
      }

      Led2 (MemoryMap.VariableRAM.Led);

      if (changeda) {
        while (ExclusionDAWrite);
///////////////////        DAControl (1, MemoryMap.VariableRAM.VariableParam.DetectLevel);
      }

      if (changebaud) {
        cli ();
        UCSR0B = 0;
        rs_hp_init (MemoryMap.VariableROM.BaudRate);
        sei ();
      }
    } else {
      if ((MemoryMap.VariableROM.StatusReturnLevel == 2) && !bcast) {
        csum = PutHeader (MemoryMap.VariableROM.ID, 2, erRange);
        PutSum (~csum);
      }
    }
  }

  uint8_t c, oldid, csum, i = 0, j = 0, len;
  volatile TPacketStatus PacketStatus = stIdle;
  static bool bcast = false, myself = false;

  InitIO ();

  rs_hp_init (MemoryMap.VariableROM.BaudRate);
///////////////////  DAControl (1, MemoryMap.VariableRAM.VariableParam.DetectLevel);

  sei ();  // 荞݋

  Led2(true);  wait(100);
  Led2(false); wait(100);
  Led2(true);  wait(100);
  Led2(false);

  MemoryMap.VariableRAM.SCANSW = 1;

  for (;;) {
    if (rs_hp_rx_buff ()){  // check rx buffer
      c = rs_hp_getc ();
      switch (PacketStatus) {
        case stIdle:
          if (c == 0xff) {
            TimeOut = 0;
            PacketStatus = stHeader;
          }
          break;
        case stHeader:
          if (c == 0xff) {
            csum = i = 0;
            PacketStatus = stLoad;
          } else {
            PacketStatus = stIdle;
          }
          break;
        case stLoad:
          csum += (packet[i] = c);
          switch (i) {
            case 0:
              if (!(c == MemoryMap.VariableROM.ID || c == BROADCASTING_ID)) {
              PacketStatus = stIdle;
              }
              break;
            case 1:
              if ((j = c - 1) + 3 >= sizeof (packet)) {
              PacketStatus = stIdle;
              }
              break;
            default:
              if (i >= sizeof (packet)) {
              PacketStatus = stIdle;
              }
              else if (--j == 0) {
                csum = ~csum;
                PacketStatus = stSum;
              }
              break;
          }
          i++;
          break;
        case stSum:
          PacketStatus = stIdle;
          if (c == csum) {
            // check ID
            bcast = (packet[0] == BROADCASTING_ID);
            myself = (packet[0] == MemoryMap.VariableROM.ID);
            // check Instruction
            switch (packet[2]) {
              case INST_READ:
                if (myself && (MemoryMap.VariableROM.StatusReturnLevel)) {
                  if (packet[3] + packet[4] <= sizeof(TMemoryMap)) {
///////////////////                    while (ExclusionMesWrite) ;
                    memcpy (&NewMemoryMap, &MemoryMap, sizeof(TMemoryMap));
                    csum = PutHeader (NewMemoryMap.VariableROM.ID, packet[4] + 2, erNon);
                    for (i = 0; i < packet[4]; i++) {
                      rs_hp_putc(NewMemoryMap.BYTE[packet[3] + i]);
                      csum += NewMemoryMap.BYTE[packet[3] + i];
                    }
                  } else {
                    csum = PutHeader (MemoryMap.VariableROM.ID, 2, erRange);
                  }
                  PutSum (~csum);
                }
                break;
              case INST_WRITE:
                if (myself || bcast) {
                  oldid = MemoryMap.VariableROM.ID;
                  len = packet[1] - 3;
                  memcpy (&NewMemoryMap, &MemoryMap, sizeof(TMemoryMap));
                  if (packet[3] + len <= sizeof (TMemoryMap) && packet[3] > 0 && packet[3] >= sizeof (TFixROM)) {
                    for (i = 0; i < len; i++) NewMemoryMap.BYTE[packet[3] + i] = packet[4 + i];
                    memwrite(oldid, bcast);
                  } else {
                    if (MemoryMap.VariableROM.StatusReturnLevel == 2 && !bcast) {
                      csum = PutHeader (MemoryMap.VariableROM.ID, 2, erRange);
                      PutSum (~csum);
                    }
                  }
                }
                break;
              case INST_SYNC_WRITE:
                if (bcast) {
                  oldid = MemoryMap.VariableROM.ID;
                  len = packet[4];
                  for (j = 0; j < (packet[1] - 4); j += (len + 1)) {
                    if (packet[5 + j] == MemoryMap.VariableROM.ID) {
                      memcpy (&NewMemoryMap, &MemoryMap, sizeof(TMemoryMap));
                      if (packet[3] + len <= sizeof (TMemoryMap) && packet[3] > 0 && packet[3] >= sizeof (TFixROM)) {
                        for (i = 0; i < len; i++) NewMemoryMap.BYTE[packet[3] + i] = packet[6 + j + i];
                        memwrite(oldid, bcast);
                      }
                      break;
                    }
                  }
                }
                break;
              case INST_PING:
                if (myself || bcast) {
                  csum = PutHeader (MemoryMap.VariableROM.ID, 2, erNon);
                  PutSum (~csum);
                }
                break;
              case INST_RESET:
                if (myself || bcast) {
                  if (MemoryMap.VariableROM.StatusReturnLevel == 2 && !bcast) {
                    csum = PutHeader (MemoryMap.VariableROM.ID, 2, erNon);
                    PutSum (~csum);
                  }
                  wait(5);
                  cli ();
                  eeprom_busy_wait ();
                  eeprom_write_block (&DefaultFixROM, 0, sizeof(TFixROM));
                  memcpy (&MemoryMap.FixROM, &DefaultFixROM, sizeof(TFixROM));
                  eeprom_busy_wait ();
                  eeprom_write_block (&DefaultVariableROM, (uint8_t *)sizeof(TFixROM), sizeof(TVariableROM));
                  memcpy (&MemoryMap.VariableROM, &DefaultVariableROM, sizeof(TVariableROM));

                  memcpy (&MemoryMap.VariableRAM.VariableParam, &MemoryMap.VariableROM.VariableParam, sizeof(TVariableParam));

                  UCSR0B = 0;
                  rs_hp_init (MemoryMap.VariableROM.BaudRate);
                  sei ();
                }
                break;
              case INST_SYS_WRITE:
                if (myself) {
                  oldid = MemoryMap.VariableROM.ID;
                  if (packet[1] >= 8) {
                    len = packet[1] - 7;
                    // R}h
                    if ((packet[3] < 128) && (len == 1)) {
                      csum = PutHeader (MemoryMap.VariableROM.ID, 2, erNon);
                      PutSum (~csum);
                      wait (10);
                      switch (packet[3]) {
                        case 0x2d:
                          cli ();
                          wdt_enable(WDTO_15MS);
                          while (1) ;
                          break;
                      }
                      break;
                    }
                    // EEPROM(0x180ȍ~)
                    else if ((packet[3] >= 128) && (packet[3] + len <= 256)) {
                      eeprom_busy_wait ();
                      eeprom_write_block ((uint8_t *)&packet[4], (uint8_t *)(0x180 + (packet[3] - 128)), len);
                      csum = PutHeader (MemoryMap.VariableROM.ID, 2, erNon);
                      PutSum (~csum);
                      break;
                    }
                  }
                  csum = PutHeader (MemoryMap.VariableROM.ID, 2, erRange);
                  PutSum (~csum);
                }
                break;
              default:
                if (MemoryMap.VariableROM.StatusReturnLevel == 2 && myself) {
                  csum = PutHeader (MemoryMap.VariableROM.ID, 2, erInst);
                  PutSum (~csum);
                }
                break;
            }
          } else{
            if (MemoryMap.VariableROM.StatusReturnLevel == 2) {
              csum = PutHeader (MemoryMap.VariableROM.ID, 2, erCheckSum);
              PutSum (~csum);
            }
          }
          break;
        default:
          PacketStatus = stIdle;
          break;
      }
    }
    // 2.048ms*1000.2048s
    if (TimeOut > (uint8_t)100U) PacketStatus = stIdle;
/*    if(MemoryMap.Name.VariableRAM.SCANSW == 1 && MemoryMap.Name.VariableROM.ID == 0 && IntervalTime == 65535) {
      Led2(0);
      FlgSYNC = (MemoryMap.Name.VariableROM.ID) ? 1 : 2;
    }
*/
  }
}
